/* PeerAcceptor - Accepts incomming connections from peers.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.net.*;
import java.io.IOException;

public class PeerAcceptor implements Runnable
{
  private final ServerSocket serverSocket;
  private final PeerCoordinator coordinator;
  private Thread thread;

  private boolean stop;

  public PeerAcceptor(int port, PeerCoordinator coordinator) throws IOException
  {
    this(new ServerSocket(port), coordinator);
  }

  public PeerAcceptor(ServerSocket serverSocket, PeerCoordinator coordinator)
    throws IOException
  {
    this.serverSocket = serverSocket;
    this.coordinator = coordinator;

    stop = false;
    thread = new Thread(this);
    thread.start();
  }

  public void stop()
  {
    stop = true;
    Thread t = thread;
    if (t != null)
      t.interrupt();
  }

  public int getPort()
  {
    return serverSocket.getLocalPort();
  }

  public void run()
  {
    boolean acceptable = true;
    while(!stop)
      {
	Socket socket = null;
	boolean exception = false;
	try
	  {
	    socket = serverSocket.accept();
	    if (coordinator.needPeers())
	      {
		Peer peer = new Peer(socket, coordinator.getID(),
				     coordinator.getMetaInfo());
		coordinator.addPeer(peer);
	      }
	    else
	      {
		socket.close();
		try
		  {
		    // Since the coordinator doesn't want more peers wait
		    // a while before accepting new ones.
		    Thread.sleep(60*1000);
		  }
		catch(InterruptedException ie) { }
	      }
	  }
	catch (IOException ioe)
	  {
	    exception = true;
	  }
	catch (Throwable t)
	  {
	    exception = true;
	    System.err.println("Exception during peer accepting: " + t);
	    t.printStackTrace();
	  }

	if (exception)
	  {
	    try
	      {
		if (socket != null) socket.close();
	      }
	    catch (IOException ignored) { }
	  }
      }
  }
}
