/* PeerConnectionOut - Keeps a queue of outgoing messages and delivers them.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.io.*;
import java.net.*;
import java.util.*;

class PeerConnectionOut implements Runnable
{
  private final Peer peer;
  private final DataOutputStream dout;

  private Thread thread;
  private boolean quit;

  // Contains Messages.
  private List sendQueue = new ArrayList();

  public PeerConnectionOut(Peer peer, DataOutputStream dout)
  {
    this.peer = peer;
    this.dout = dout;

    quit = false;
    thread = new Thread(this);
    thread.start();
  }

  /**
   * Continuesly monitors for more outgoing messages that have to be send.
   * Stops if quit is true of an IOException occurs.
   */
  public void run()
  {
    try
      {
	while (!quit)
	  {
	    Message m = null;
	    synchronized(sendQueue)
	      {
		while (!quit && sendQueue.isEmpty())
		  {
		    try
		      {
			// Make sure everything will reach to other side.
			dout.flush();
			
			// Wait till more data arrives.
			sendQueue.wait();
		      }
		    catch (InterruptedException ie)
		      {
			/* ignored */
		      }
		  }
		if (!quit)
		  m = (Message)sendQueue.remove(0);
	      }
	    if (m != null)
	      {
		m.sendMessage(dout);
		// XXX - Should also register overhead...
		if (m.type == Message.PIECE)
		  peer.state.uploaded(m.len);
	      }
	  }
      }
    catch (IOException ioe)
      {
	// Ignore, probably other side closed connection.
      }
    catch (Throwable t)
      {
	System.err.println(peer + ": "  + t);
	t.printStackTrace();
      }
    finally
      {
	quit = true;
	peer.disconnect();
      }
  }

  public void disconnect()
  {
    synchronized (sendQueue)
      {
	if (quit == true)
	  return;
	
	quit = true;
	thread.interrupt();
	
	try
	  {
	    dout.close();
	  }
	catch (IOException ioe)
	  {
	    /* ignored */
	  }

	sendQueue.clear();
	sendQueue.notify();
      }
  }

  /**
   * Adds a message to the sendQueue and notifies the method waiting
   * on the sendQueue to change.
   */
  private void addMessage(Message m)
  {
    synchronized(sendQueue)
      {
	sendQueue.add(m);
	sendQueue.notify();
      }
  }

  void sendAlive()
  {
    Message m = new Message();
    m.type = Message.KEEP_ALIVE;
    addMessage(m);
  }

  void sendChoke(boolean choke)
  {
    Message m = new Message();
    if (choke)
      m.type = Message.CHOKE;
    else
      m.type = Message.UNCHOKE;
    addMessage(m);
  }

  void sendInterest(boolean interest)
  {
    Message m = new Message();
    if (interest)
      m.type = Message.INTERESTED;
    else
      m.type = Message.UNINTERESTED;
    addMessage(m);
  }

  void sendHave(int piece)
  {
    Message m = new Message();
    m.type = Message.HAVE;
    m.piece = piece;
    addMessage(m);
  }

  void sendBitfield(BitField bitfield)
  {
    Message m = new Message();
    m.type = Message.BITFIELD;
    m.data = bitfield.getFieldBytes();
    m.off = 0;
    m.len = m.data.length;
    addMessage(m);
  }

  void sendRequests(List requests)
  {
    Iterator it = requests.iterator();
    while (it.hasNext())
      {
	Request req = (Request)it.next();
	sendRequest(req);
      }
  }

  void sendRequest(Request req)
  {
    Message m = new Message();
    m.type = Message.REQUEST;
    m.piece = req.piece;
    m.begin = req.off;
    m.length = req.len;
    addMessage(m);
  }

  synchronized void sendPiece(int piece, int begin, int length, byte[] bytes)
  {
    Message m = new Message();
    m.type = Message.PIECE;
    m.piece = piece;
    m.begin = begin;
    m.length = length;
    m.data = bytes;
    m.off = begin;
    m.len = length;
    addMessage(m);
  }

  synchronized void sendCancel(int piece, int begin, int length)
  {
    Message m = new Message();
    m.type = Message.CANCEL;
    m.piece = piece;
    m.begin = begin;
    m.length = length;
    addMessage(m);
  }
}
